#!/usr/bin/env -S python3 -ttu

# legonigma.py
# Samsung CTF 2022 - Legonigma
# Copyright 2022, mal@sec.gd
# Released under EUPLv1.2

from __future__ import annotations

import logging
from math import ceil, floor
from typing import List, Tuple

logging.basicConfig(level=logging.DEBUG)
LOGGER = logging.getLogger()

CIPHERTEXT = "59x8wl9pjsava3grn0il79aoq307f20a6huc3xnos289cd8xv1fn2znuoa2bq8959chbktwfow8c3azpvo3c59jz4"

MAP = "abcdefghijklmnopqrstuvwxyz0123456789"


def letter_to_rotation(letter: str) -> float:
    return MAP.index(letter) / len(MAP)


def rotation_to_letter(rotation: float) -> str:
    idx_float = (rotation % 1) * len(MAP)
    letter = MAP[round(idx_float) % len(MAP)]
    if idx_float % 1 > 0.0001 and idx_float % 1 < 0.9999:
        LOGGER.warning("Rounding %r to %r", idx_float, letter)
    return letter


class Gadget:
    in_state: float
    """The current rotation count of the input"""

    def __init__(self, in_state: float | str = 0.0):
        if isinstance(in_state, float):
            self.in_state = in_state
            return
        self.in_state = 0.0
        for letter in in_state:
            self.rotate_to_letter(letter)

    def rotate_to_letter(self, letter: str):
        # LOGGER.debug("In_was: %r (%r) to %r", self.in_state, rotation_to_letter(self.in_state), letter)
        letter_rotation = letter_to_rotation(letter)
        full_rotations = self.in_state // 1
        if self.in_state % 1 >= letter_rotation:
            full_rotations += 1
        self.in_state = full_rotations + letter_rotation
        # LOGGER.debug("In: %r (%r)", self.in_state, rotation_to_letter(self.in_state))

    def read(self, *offsets: List[int, ...]) -> Tuple[str, ...]:
        if not offsets:
            offsets = (0, 0, 0)
        # ratios = (5,7,13)
        # ratios = (5,13,7)
        # ratios = (7,5,13)
        # ratios = (7,13,5)
        ratios = (13, 5, 7)
        # ratios = (13,7,5)
        nums = [
            self.in_state * ratios[i] + offsets[i] / len(MAP)
            for i in range(len(ratios))
        ]
        outputs = [rotation_to_letter(n) for n in nums]
        # print(ratios, offsets, nums, outputs)
        return outputs

    def advance(self, n: int = 1):
        self.in_state += letter_to_rotation("b")


# for i in MAP:
#    test = ""+i
#    r = Gadget(test).read()
#    print(test, " ".join(r))
#    n = sum((1 for x in r if x in "x"))
#    if n >= 1:
#        print("MAYBE???????????????????", n)

n = 0
plaintext = ""
for c in CIPHERTEXT:
    for p in MAP:
        cc = Gadget(p).read()[n]
        if cc == c:
            plaintext += p
    n = (n + 1) % 3

print(plaintext)


# expected_start = "59x"
# for i in MAP:
#    r = Gadget(i).read()
#    offsets = [(MAP.index(expected_start[j]) - MAP.index(r[j])) % len(MAP) for j in range(3)]
#    print(offsets)
#    for j in MAP:
#        r = Gadget(j).read(*offsets)
#        print(i, j, " ".join(r))
#        n = sum((1 for x in r if x in ("8", "w", "l")))
#        if n >= 3:
#            print("MAYBE???????????????????", n)
#            exit(0)

# g = Gadget()
# for i in range(500000):
#    g.advance()
#    r = g.read()
#    print(g.in_state, r)
#    n = sum((1 for x in r if x in ("5", "9", "x")))
#    if n >= 2:
#        print("MAYBE???????????????????", n)
#        #break

# for rot in range(36):
#    rotmap = "".join([MAP[(MAP.index(i)+rot)%len(MAP)] for i in MAP])
#    print(rot, rotmap)

# d={i: Gadget(i).read()[0] for i in MAP}
# rd={v: k for k, v in d.items()}
#
# for rot in range(36):
#    pt="".join([rd[
#            MAP[(MAP.index(j)+rot)%len(MAP)]
#        ] for j in CIPHERTEXT])
#    for i in range(3):
#        print(pt[i::3])
